#include "ecc.h"
#include <stdio.h>
#include <stdlib.h>


// points always to the currently used generator polynom
static uint8_t *current_gen_poly;
// leading power of the current used generator polynom
static uint8_t current_gen_poly_deg;
// stores the number of payload bytes that can be encoded per block with the current BCH enc mode
static uint8_t current_payload_bytes;

// holds the buffers and sizes for encoding/decoding
static char* inBuf;
static char* outBuf;

//  for byte to bit conversion (used in Hamming, DECTED)
static uint8_t dataBits[16];
//  for bit to byte conversion (used in Hamming, DECTED)
static uint8_t enc_dataBits[16];

// stores the num of errors that can be corrected using BCH modes (corresponds to t)
static uint8_t current_err_capability;

#define UPDIV(x,y) (x)/(y)+((x)%(y)?1:0)

// holds sizes
static uint8_t inBuf_s;
static uint8_t outBuf_s;

static bool isEncoding;

enum ECC_mode {
	OFF 		= 0x00,		///< No FEC 									=  0
	HAMMING74	= 0x01,		///< Hamming Code (7,4) SECDED					=  2
	DECTED168	= 0x02,		///< pseudo Hamming Code (16,8) DECTED			=  3
	BCH6345		= 0x03, 	///< BCH(63,45) corrects 3-bit error 			=  6
	BCH6339		= 0x04, 	///< BCH(63,39) corrects 4-bit error 			=  7
	BCH6336		= 0x05, 	///< BCH(63,36) corrects 5-bit error 			=  8
	BCH6357		= 0x06, 	///< BCH(63,57) corrects 1-bit error 			=  4
	BCH6351		= 0x07, 	///< BCH(63,51) corrects 2-bit error 			=  5
	REP38		= 0x08		///< Repetition Code (3,8)						=  1
};

static char const *ecc_mode[]={
				"OFF",
				"HAMMING74",
				"DECTED168",
				"BCH6345",
				"BCH6339",
				"BCH6336",
				"BCH6357",
				"BCH6351",
				"REP38"
			 };

typedef struct {
  enum ECC_mode		current_ECC_mode;
  uint32_t			last_ECC_packet;
  bool				ECC_senderOn;
  uint8_t 			rcv_node;
  size_t			payloadSize;
} ecc_config_t;

static ecc_config_t	ECCConfiguration;

// holds statistics about corrected errors
struct _ECC_err_stats {
	uint8_t max;
	uint8_t min;
	int err_sum; ///> avg error per block (ceiled)
	int block_counter;
} ECC_corrected_error_stats;


struct _ECC_decoder_state {
	bool CORRECTED_ERROR;
	bool DETECTED_ERROR;
	bool DECODER_SUCCESSFUL;
} ECC_decoder_state;


// holds the parameters for hamming code
struct param {
      uint8_t n;	// length of encoded bit sequence
      uint8_t k;	// length of data
      uint8_t r;	// length of redundancy (ham.n-ham.k)
};

// paramters for code Hamming(7,4);
static struct param ham;
static struct param bch;

static char enc_buf[ECC_MAX_PAYLOAD_SIZE];
static char dec_buf[ECC_MAX_DEC_PAYLOAD_SIZE];


/*************************************************************************************************************************/
/*  FUNCTIONS                  *******************************************************************************************/
/*************************************************************************************************************************/



#define TEST_INPUT_SIZE 	20


void ECC_run_test_program() {
	printf("ERROR CORRECTION CODES TEST PROGRAM\n\n");

	uint8_t current_mode;

	for(current_mode=OFF; current_mode<=BCH6351; current_mode++) {

		const char* label = ecc_mode[current_mode];
		printf("testing current ecc mode: %s\n", label);

		ECC_setMode(current_mode);

		memset((void *) &enc_buf, 0, ECC_MAX_PAYLOAD_SIZE);
		memset((void *) &dec_buf, 0, ECC_MAX_DEC_PAYLOAD_SIZE);

		char originaldata_buffer[TEST_INPUT_SIZE];
		memset((void *) &originaldata_buffer, 0, TEST_INPUT_SIZE);
		sprintf(originaldata_buffer, "%s", "TEST PAYLOAD!");

		uint16_t enc_bufSize = ECC_calcBufferSize(TEST_INPUT_SIZE, true);
		printf("enc_bufSize %i\n", enc_bufSize);

		bool encodeSuccess = ECC_encode((char*)originaldata_buffer, (uint16_t)TEST_INPUT_SIZE, (char *) enc_buf, (uint16_t)enc_bufSize);
		if (encodeSuccess) {
			printf("encoding successful\n");
		} else {
			printf("encoding failed\n");
		}

		uint16_t numOfBytes;
		uint16_t dec_bufSize = ECC_calcBufferSize(enc_bufSize, false);
		printf("dec_bufSize %i\n", dec_bufSize);

		numOfBytes = ECC_decode((char*)enc_buf, (uint16_t) enc_bufSize, (char*) dec_buf, (uint16_t) dec_bufSize);
		printf("decoded: %s\n", dec_buf);

		bool decSuccess = true;
		int pos;
		for (pos = 0; pos < numOfBytes; pos++) {
			if (dec_buf[pos] != originaldata_buffer[pos]) {
				decSuccess = false;
				break;
			}
		}
		if (decSuccess) {
			printf("decoding successful\n\n");
		} else {
			printf("decoding failed\n\n");
		}
	}
}


void ECC_run_test_program_II() {

	printf("ERROR CORRECTION CODES TEST PROGRAM\n");

	ECC_setMode(BCH6351);

    char orig_data[] = {87, 84, 82, 94, 70, 118, 22, 214, 10, 20};

    uint8_t size_inBuf, size_outBuf;
    size_inBuf = ECC_calcBufferSize(sizeof(orig_data), true);
    size_outBuf = ECC_calcBufferSize(size_inBuf, false);


    char enc_data[size_inBuf];
    char dec_data[size_outBuf];

    int l;

    for (l = 0; l < 1; l++) {

	        printf("original data\n");
	        ECC_printBytesRow(orig_data, sizeof(orig_data));

	        bool encoded;
	        encoded = ECC_encode(orig_data, sizeof(orig_data), enc_data, size_inBuf);
	        printf("encoded data\n");
	        ECC_printBytesRow(enc_data, size_inBuf);

	        uint8_t numOfBytes = ECC_decode(enc_data, size_inBuf, dec_data, size_outBuf);
	        printf("decoded data (len: %d)\n", numOfBytes);
	        ECC_printBytesRow(dec_data, numOfBytes);

	        ECC_printErrStats();

	        bool decSuccess = true;
	        int pos;
	        for (pos = 0; pos < numOfBytes; pos++) {
	            if (dec_data[pos] != orig_data[pos]) {
	                decSuccess = false;
	                break;
	            }
	        }
	        if (decSuccess) {
	            printf("decoding successful\n");
	        } else {
	            printf("decoding failed\n");
	        }
	    }
}

void ECC_printErrStats() {
	printf("min_err: %u\t", ECC_corrected_error_stats.min);
	printf("max_err: %u\t", ECC_corrected_error_stats.max);
	printf("blocks: %u\t", ECC_corrected_error_stats.block_counter);
	printf("sum_err: %u\t", ECC_corrected_error_stats.err_sum);
}

void ECC_printBytesRow(char* buf, uint8_t buf_s) {

	uint8_t j;
	// pos represents 1 0 0 0 0 0 0 0 which is shifted to right
	uint8_t pos;
	uint8_t bit_mask;

	if (buf_s > 0) {
		printf("|  ");
		for (pos = buf_s; pos > 0; pos--) {
			printf("%p", &(buf[pos - 1]));
			printf("  |  ");
		}
		printf("\n");
		printf("| ");
		for (pos = buf_s; pos > 0; pos--) {
			bit_mask = 0x80;
			for (j = 0; j < 8; j++) {
				if ((buf[pos - 1] & bit_mask) == bit_mask){
					printf("1");
				} else {
					printf("0");
				}
				// shift pos to right
				bit_mask = (bit_mask >> 1);
			}
			printf(" | ");
		}
	}
	printf("\n");
}

void ECC_setMode(uint8_t mode) {

	switch (mode) {
			case OFF:
				ECCConfiguration.current_ECC_mode = OFF;
				ham.n = 0;
				ham.k = 0;
				ham.r = ham.n - ham.k;
				current_gen_poly = NULL;
				current_gen_poly_deg = 0;
				current_payload_bytes = 0;
				current_err_capability = 0;
				bch.n = 0;
				bch.k = 0;
				break;
			case HAMMING74:
				ECCConfiguration.current_ECC_mode = HAMMING74;
				current_err_capability = 1;
				current_gen_poly = NULL;
				current_gen_poly_deg = 0;
				current_payload_bytes = 0;
				ham.n = 7;
				ham.k = 4;
				ham.r = ham.n - ham.k;
				bch.n = 0;
				bch.k = 0;
				break;
			case DECTED168:
				ECCConfiguration.current_ECC_mode = DECTED168;
				current_err_capability = 2;
				current_gen_poly = NULL;
				current_gen_poly_deg = 0;
				current_payload_bytes = 0;
				ham.n = 16;
				ham.k = 8;
				ham.r = ham.n - ham.k;
				bch.n = 0;
				bch.k = 0;
				break;
			case BCH6345: // BCH6345
				ECCConfiguration.current_ECC_mode = BCH6345;
				current_err_capability = 3;
				current_gen_poly = gen_poly_BCH_63_45;
				current_payload_bytes = 5;
				ham.n = 0;
				ham.k = 0;
				ham.r = ham.n - ham.k;
				bch.n = 63;
				bch.k = 45;
				current_gen_poly_deg = bch.n - bch.k;
				break;
			case BCH6339: // BCH6339
				ECCConfiguration.current_ECC_mode = BCH6339;
				current_err_capability = 4;
				current_gen_poly = gen_poly_BCH_63_39;
				current_payload_bytes = 4;
				ham.n = 0;
				ham.k = 0;
				ham.r = ham.n - ham.k;
				bch.n = 63;
				bch.k = 39;
				current_gen_poly_deg = bch.n - bch.k;
				break;
			case BCH6336: // BCH6336
				ECCConfiguration.current_ECC_mode = BCH6336;
				current_err_capability = 5;
				current_gen_poly = gen_poly_BCH_63_36;
				current_payload_bytes = 4;
				ham.n = 0;
				ham.k = 0;
				ham.r = ham.n - ham.k;
				bch.n = 63;
				bch.k = 36;
				current_gen_poly_deg = bch.n - bch.k;
				break;
			case BCH6357: // BCH6357
				ECCConfiguration.current_ECC_mode = BCH6357;
				current_err_capability = 1;
				current_gen_poly = gen_poly_BCH_63_57;
				current_payload_bytes = 7;
				ham.n = 0;
				ham.k = 0;
				ham.r = ham.n - ham.k;
				bch.n = 63;
				bch.k = 57;
				current_gen_poly_deg = bch.n - bch.k;
				break;
			case BCH6351: // BCH6351
				ECCConfiguration.current_ECC_mode = BCH6351;
				current_err_capability = 2;
				current_gen_poly = gen_poly_BCH_63_51;
				current_payload_bytes = 6;
				ham.n = 0;
				ham.k = 0;
				ham.r = ham.n - ham.k;
				bch.n = 63;
				bch.k = 51;
				current_gen_poly_deg = bch.n - bch.k;
				break;
			case REP38:
				ECCConfiguration.current_ECC_mode = REP38;
				current_err_capability = 1;
				ham.n = 0;
				ham.k = 0;
				ham.r = ham.n - ham.k;
				bch.n = 0;
				bch.k = 0;
				current_gen_poly = NULL;
				current_gen_poly_deg = 0;
				current_payload_bytes = 0;
				break;
			default:
				ECCConfiguration.current_ECC_mode = OFF;
				ham.n = 0;
				ham.k = 0;
				ham.r = ham.n - ham.k;
				current_gen_poly = NULL;
				current_gen_poly_deg = 0;
				current_payload_bytes = 0;
				current_err_capability = 0;
				bch.n = 0;
				bch.k = 0;
				break;
		}
}


uint8_t ECC_calcBufferSize(uint8_t size, bool isEncode) {

	uint8_t newSize = 0;
	float temp;
	switch (ECCConfiguration.current_ECC_mode) {
		case OFF: // No FEC
			newSize = (uint16_t) size;
		break;
		case HAMMING74: // Hamming74
			if (isEncode) {
				temp = ((float) size) * 14.0 / 8.0;
				if ( (int) (10 * temp) % 10 != 0) {
					newSize = (uint8_t) (temp + 1);
				} else {
					newSize = (uint8_t) temp;
				}
			} else {
				newSize = (uint8_t) ((float) size / 14.0 * 8.0);
			}
			break;
		case DECTED168: // DECTED (16,8)
			if (isEncode) {
				newSize = size * 2;
			} else {
				newSize = size / 2;
			}
			break;
		case BCH6345: // BCH6345
			if (isEncode) {
				temp = ((float) size) / 5.0;
				if ( (uint8_t) (10 * temp) % 10 != 0) {
					newSize = (uint8_t) (temp + 1);
				} else {
					newSize = (uint8_t) temp;
				}
				// it uses the one bit left too.
				newSize *= 8;
			} else {
				// decoded size is mostly too high need payload size field in payload struct to show the end of the payload
				newSize = (uint8_t) ((float) size / 64.0 * 45.0);
			}
			break;
		case BCH6339: // BCH6339
			if (isEncode) {
				temp = ((float) size) / 4.0;
				if ( (uint8_t) (10 * temp) % 10 != 0) {
					newSize = (uint8_t) (temp + 1);
				} else {
					newSize = (uint8_t) temp;
				}
				// it uses the one bit left too.
				newSize *= 8;
			} else {
				// decoded size is mostly too high need payload size field in payload struct to show the end of the payload
				newSize = (uint8_t) ((float) size / 64.0 * 39.0);
			}
			break;
		case BCH6336: // BCH6336
			if (isEncode) {
				temp = ((float) size) / 4.0;
				if ( (uint8_t) (10 * temp) % 10 != 0) {
					newSize = (uint8_t) (temp + 1);
				} else {
					newSize = (uint8_t) temp;
				}
				// it uses the one bit left too.
				newSize *= 8;
			} else {
				// decoded size is mostly too high need payload size field in payload struct to show the end of the payload
				newSize = (uint8_t) ((float) size / 64.0 * 36.0);
			}
			break;
		case BCH6357: // BCH6357
			if (isEncode) {
				temp = ((float) size) / 7.0;
				if ( (int) (10 * temp) % 10 != 0) {
					newSize = (uint8_t) (temp + 1);
				} else {
					newSize = (uint8_t) temp;
				}
				// it uses the one bit left too.
				newSize *= 8;
			} else {
				// decoded size is mostly too high need payload size field in payload struct to show the end of the payload
				newSize = (uint8_t) ((float) size / 64.0 * 57.0);
			}
			break;
		case BCH6351: // BCH6351
			if (isEncode) {
				temp = ((float) size) / 6.0;
				if ( (uint8_t) (10 * temp) % 10 != 0) {
					newSize = (uint8_t) (temp + 1);
				} else {
					newSize = (uint8_t) temp;
				}
				// it uses the one bit left too.
				newSize *= 8;
			} else {
				// decoded size is mostly too high need payload size field in payload struct to show the end of the payload
				newSize = (uint8_t) ((float) size / 64.0 * 51.0);
			}
			break;
		case REP38: // REP38
			if (isEncode) {
				newSize = size * 3;
			} else {
				newSize = UPDIV((int) size, 3);
			}
			break;
		default: // No FEC
			newSize = (uint16_t) size;
			break;
	}
	return newSize;
}



bool ECC_encode(char* in, uint16_t in_s, char* out, uint16_t out_s) {

	// check if encoded size is too big
	if (out_s > ECC_MAX_PAYLOAD_SIZE) {
		return false;
	}

	// holds state about encoding
	bool encSuccess;

	uint8_t pos;

	// assign buffers and sizes to global vars
	inBuf = in;
	inBuf_s = in_s;

	outBuf = out;
	outBuf_s = out_s;

	switch (ECCConfiguration.current_ECC_mode) {
		case OFF:
			if (inBuf_s == outBuf_s) {
				for (pos = 0; pos < outBuf_s; pos++) {
					outBuf[pos] = inBuf[pos];
				}
				encSuccess = true;
			} else {
				encSuccess = false;
			}
			break;

		case HAMMING74: // Hamming (7,4)
			encSuccess = ECC_enc_ham();
			break;

		case DECTED168: // DECTED (16,8)
			encSuccess = ECC_enc_DECTED();
			break;

		case BCH6345: // BCH6345
			encSuccess = ECC_enc_bch();
			break;
		case BCH6339: // BCH6339
			encSuccess = ECC_enc_bch();
			break;
		case BCH6336: // BCH6336
			encSuccess = ECC_enc_bch();
			break;
		case BCH6357: // BCH6357
			encSuccess = ECC_enc_bch();
			break;
		case BCH6351: // BCH6351
			encSuccess = ECC_enc_bch();
			break;
		case REP38: // REP38
			encSuccess = ECC_enc_rep(3, 8);
			break;

		default: // No FEC - just pass through buffer content
			if (inBuf_s == outBuf_s) {
				for (pos = 0; pos < outBuf_s; pos++) {
					outBuf[pos] = inBuf[pos];
				}
				encSuccess = true;
			} else {
				encSuccess = false;
			}
			break;
	}
	return encSuccess;
}


bool ECC_enc_rep(uint8_t r, uint8_t k) {
	// protect encoding process
	isEncoding = true;

	if ((r % 2 == 0) && r < 2)  {
		// check for r to be an odd num bigger than one otherwise majority decision is not unique
		return false;
	}

	uint8_t i, j, outBuf_pos;
	outBuf_pos = 0;
	for (i = 0; i < inBuf_s; i++) {

		for (j = 0; j < r; j++) {
			if (outBuf_pos < outBuf_s) {
				// fill new buffer
				outBuf[outBuf_pos] = inBuf[i];
				// print the bytes
				outBuf_pos++;
			} else {
				printf("Out buffer size mismatch\n");
				return false;
			}
		}
	}
	// protect encoding process finished
	isEncoding = false;

	return true;
}



bool ECC_enc_ham() {

	uint16_t num_dataBits;
	num_dataBits = ((uint16_t) inBuf_s) * 8;

	uint16_t num_encDataBits;
	num_encDataBits = ((uint16_t) outBuf_s) * 8;

	if ((num_encDataBits <= ((uint16_t) ECC_MAX_PAYLOAD_SIZE * (uint16_t) 8)) &&
				(num_dataBits <= ((uint16_t) ECC_MAX_DEC_PAYLOAD_SIZE * (uint16_t) 8))) {

		ECC_doHammingEncoding(inBuf, inBuf_s, outBuf, outBuf_s);

		return true;
	} else {
		printf("data for encoding too large\t");
		return false;
	}
}


void ECC_doHammingEncoding(char* in, uint16_t in_s, char* out, uint16_t out_s) {

	// protect encoding process
	isEncoding = true;

	uint16_t s;
	uint8_t dataBlock[ham.k];
	uint8_t enc_dataBits[ham.n];

	uint8_t byte;
	uint8_t bit;
	uint8_t bit_counter1, bit_counter2;
	char offset1, offset2;

	// init the params;
	bit_counter1 = 1;
	bit_counter2 = 1;
	byte = 0;
	memset((void*) out, 0, out_s);	// reset all bits in the bit stream (since there are not filled bits at the end)
	offset2 = 0x01;
	offset1 = 0x01;
	// create the data blocks of size k of input data
	for (s = 0; s < in_s; ) {
		// read k-bits
		for (bit = 0; bit < ham.k; bit++) {
			// get the bits from the bytes
			if ((in[s] & offset1) == offset1) {
				dataBlock[bit] = 1;
			} else {
				dataBlock[bit] = 0;
			}

			if (bit_counter1 == 8) {
				offset1 = 0x01;
				s++;
				bit_counter1 = 1;
			} else {
				offset1 = offset1 << 1;
				bit_counter1++;
			}
		}

		//reset encoded bits
		memset((uint8_t*) &enc_dataBits, 0, ham.n);


		if (ECCConfiguration.current_ECC_mode == HAMMING74) {
			ECC_mult_with_mat((uint8_t*) &dataBlock, ham.k, (uint8_t*) &generator_ham74, (uint8_t*) &enc_dataBits, ham.n);
		} else if (ECCConfiguration.current_ECC_mode == DECTED168) {
			ECC_mult_with_mat((uint8_t*) &dataBlock, ham.k, (uint8_t*) &generator_ham168, (uint8_t*) &enc_dataBits, ham.n);
		}
		// create the bytes
		uint8_t e;
		for (e = 0; e < ham.n; e++) {
			// write bit in out buffer
			if (enc_dataBits[e] == 1) {
				out[byte] = out[byte] | offset2;
			} else if (enc_dataBits[e] == 0){
				out[byte] = out[byte] & ~offset2;
			} else {
				printf("error in bit reading\n");
			}
			if (bit_counter2 == 8) {
				byte++;
				bit_counter2 = 1;
				offset2 = 0x01;
			} else {
				offset2 = offset2 << 1;
				bit_counter2++;
			}
		}
	}

	// protect encoding process finished
	isEncoding = false;
}

void ECC_mult_with_mat(uint8_t* input_vec, uint8_t input_vec_s, uint8_t* matrix, uint8_t* output_vec, uint8_t output_vec_s) {
	uint8_t inPos;
	uint8_t matPos;
	uint8_t outPos;
	uint8_t matCol;
	matCol = 0;
	matPos = 0;
	for (outPos = 0; outPos < output_vec_s; outPos++) {
		matPos = matCol;
		output_vec[outPos] = 0;
		for (inPos = 0; inPos < input_vec_s; inPos++) {
			output_vec[outPos] = (output_vec[outPos] + (input_vec[inPos] * matrix[matPos])) % 2;
			matPos += output_vec_s;
		}
		matCol++;
	}
}

bool ECC_enc_DECTED() {

	uint16_t num_dataBits;
	num_dataBits = ((uint16_t) inBuf_s) * 8;

	uint16_t num_encDataBits;
	num_encDataBits = ((uint16_t) outBuf_s) * 8;

	if ((num_encDataBits <= ((uint16_t) ECC_MAX_PAYLOAD_SIZE * (uint16_t) 8)) &&
					(num_dataBits <= ((uint16_t) ECC_MAX_DEC_PAYLOAD_SIZE * (uint16_t) 8))) {

		ECC_doHammingEncoding(inBuf, inBuf_s, outBuf, outBuf_s);
		return true;
	} else {
		printf("data for encoding too large\t");
		return false;
	}
}


/*************************************************************************************************************************/
/* DECODING PART																										 */
/*************************************************************************************************************************/

uint8_t ECC_decode(char* in, uint16_t in_s, char* out, uint16_t out_s) {
	// holds the number of bytes to read for the application
	uint8_t numOfPayloadBytes = 0;
	// reset the error stats
	ECC_initErrStats();
	// if no ECC is disable just copy inbuffer to outbuffer
	uint8_t pos;

	// assign buffers and sizes to global vars
	inBuf = in;
	inBuf_s = in_s;

	outBuf = out;
	outBuf_s = out_s;

	switch (ECCConfiguration.current_ECC_mode) {
		case OFF: // No FEC
			if (inBuf_s == outBuf_s) {
				for (pos = 0; pos < outBuf_s; pos++) {
					outBuf[pos] = inBuf[pos];
				}
				numOfPayloadBytes = inBuf_s;
			} else {
				printf("size of buffers do not match (size: %u)\n", inBuf_s);
			}
			// since no error correction is done
			ECC_setErrStats(0);
			break;

		case HAMMING74: // Hamming (7,4)
			numOfPayloadBytes = ECC_dec_ham();
			break;

		case DECTED168: // DECTED (16,8)
			numOfPayloadBytes = ECC_dec_DECTED();
			break;
		case BCH6345: // BCH6345
			numOfPayloadBytes = ECC_dec_bch();
			break;
		case BCH6339: // BCH6339
			numOfPayloadBytes = ECC_dec_bch();
			break;
		case BCH6336: // BCH6336
			numOfPayloadBytes = ECC_dec_bch();
			break;
		case BCH6357: // BCH6357
			numOfPayloadBytes = ECC_dec_bch();
			break;
		case BCH6351: // BCH6351
			numOfPayloadBytes = ECC_dec_bch();
			break;
		case REP38: // REP38
			numOfPayloadBytes = ECC_dec_rep(3, 8);
			break;

		default: // No FEC
			if (inBuf_s == outBuf_s) {
				for (pos = 0; pos < outBuf_s; pos++) {
					outBuf[pos] = inBuf[pos];
				}
				numOfPayloadBytes = inBuf_s;
			}
			// since no error correction is done
			ECC_setErrStats(0);
			break;
	}
	if (numOfPayloadBytes == 0) {
		printf("size error while decoding\t");
	}

	return numOfPayloadBytes;
}


void ECC_initErrStats() {
	// init or reset values
	ECC_corrected_error_stats.max = 0;
	ECC_corrected_error_stats.min = 255;
	ECC_corrected_error_stats.block_counter = 0;
	ECC_corrected_error_stats.err_sum = 0;
}

uint8_t ECC_dec_ham() {

	uint16_t num_encBits;
	num_encBits = ((uint16_t) inBuf_s) * (uint16_t) 8;

	uint16_t num_decDataBits;
	num_decDataBits = ((uint16_t) outBuf_s) * (uint16_t) 8;

	uint8_t num_decDataBytes;
	num_decDataBytes = outBuf_s;

	if ((num_encBits <= ((uint16_t) ECC_MAX_PAYLOAD_SIZE * (uint16_t) 8)) &&
			(num_decDataBits <= ((uint16_t) ECC_MAX_DEC_PAYLOAD_SIZE * (uint16_t) 8))) {

		ECC_doHammingDecoding(inBuf, inBuf_s, outBuf, outBuf_s);

	} else {
		printf("data for decoding too large\t");
	}
	return num_decDataBytes;
}



void ECC_doHammingDecoding(char* in, uint16_t in_s, char* out, uint16_t out_s) {
	uint16_t s;
	uint8_t dataBlock[ham.n];
	uint8_t syn[ham.r];
	uint8_t error_pos;
	bool conv_err;

	uint8_t byte;
	uint8_t bit;
	uint8_t bit_counter1, bit_counter2;
	char offset1, offset2;

	// init the params
	bit_counter1 = 1;
	bit_counter2 = 1;
	byte = 0;
	memset((void*) out, 0, out_s);	// reset all bits in the bit stream (since there are not filled bits at the end)
	offset2 = 0x01;
	offset1 = 0x01;

	// create the data blocks of size k of input data
	for (s = 0; s < in_s; ) {
		if(s == in_s -1) {
			break;
		}
		// read the n-bits to decode
		for (bit = 0; bit < ham.n; bit++) {

			// get the bits from the bytes
			if ((in[s] & offset1) == offset1) {
				dataBlock[bit] = 1;
			} else {
				dataBlock[bit] = 0;
			}
			// check if next byte begins
			if (bit_counter1 == 8) {
				// reset offset since reading from new byte
				offset1 = 0x01;
				// begin to read next byte
				s++;
				// count bits from the next byte
				bit_counter1 = 1;
			} else {
				// still in current byte
				offset1 = offset1 << 1;
				bit_counter1++;
			}
		}

		// reset syndrom array
		memset((uint8_t*) &syn, 0, ham.r);

		// start the decoding
		if (ECCConfiguration.current_ECC_mode == HAMMING74) {
			// calculate syndrom value
			ECC_mult_with_mat((uint8_t*) &dataBlock, ham.n, (uint8_t*) &parityCheck_ham74, (uint8_t*) &syn, ham.r);
			conv_err = false;
			error_pos = 0;
			// calculate the bit error pos
			conv_err = ECC_convertBit2Byte((uint8_t*) &syn, ham.r, &error_pos);
			if ((error_pos > 0) && (conv_err == false)) {
				// do the hamming error detection and correction
				ECC_doHammingCorrection((uint8_t*) &dataBlock, (uint8_t*) &syn);
			} else {
				// fill the err stat
				if (COUNT_ERR == true) {
					// found error in the block
					ECC_setErrStats(0);
				}
			}
		} else if (ECCConfiguration.current_ECC_mode == DECTED168) {
			// calculate syndrom value
			ECC_mult_with_mat((uint8_t*) &dataBlock, ham.n, (uint8_t*) &parityCheck_ham168, (uint8_t*) &syn, ham.r);
			conv_err = false;
			error_pos = 0;
			// calculate the bit error pos
			conv_err = ECC_convertBit2Byte((uint8_t*) &syn, ham.r, &error_pos);
			if ((error_pos > 0) && (conv_err == false)) {
				// do the DECTED Correction
				ECC_doDECTEDCorrection((uint8_t*) &dataBlock, ham.n, (uint8_t*) &syn);
			} else {
				// fill the err stat
				if (COUNT_ERR == true) {
					// found error in the block
					ECC_setErrStats(0);
				}
			}
		}
		// create the bytes

		//printf("corrected block\n");
		// ECC_printBytesRow((uint8_t*) &dataBlock, ham.n);

		// printf("\ndec:\n");
		uint8_t e;
		for (e = 0; e < ham.k; e++) {
			// write bit in out buffer
			if (dataBlock[e] == 1) {
				out[byte] = out[byte] | offset2;
				// printf("1");
			} else if (dataBlock[e] == 0){
				out[byte] = out[byte] & ~offset2;
				// printf("0");
			} else {
				printf("error in bit reading\n");
			}
			if (bit_counter2 == 8) {
				byte++;
				bit_counter2 = 1;
				offset2 = 0x01;
			} else {
				offset2 = offset2 << 1;
				bit_counter2++;
			}
		}
	}
}

void ECC_doHammingCorrection(uint8_t* data, uint8_t* syn) {

	uint8_t pos;
	bool conv_err;
	bool temp_conv_err;
	uint8_t syn_byte;
	uint8_t mat_pos;
	pos = 0;
	conv_err = ECC_convertBit2Byte(syn, ham.r, &pos);

	if ((pos > 0) && (pos < 8) && (conv_err == false)) {
		// go through rows to find erroneous position
		uint8_t row = 0;
		// check which row corresponds to the syndrome. the row corresponds to the erroneus pos.
		uint8_t t;
		for (t = 0; t < ham.n * (ham.r); t += (ham.r)) {
			syn_byte = 0;
			mat_pos = 0;
			// in order to overcome lazy evaluation
			temp_conv_err = ECC_convertBit2Byte((uint8_t*) &parityCheck_ham74[t], ham.r, &mat_pos);
			conv_err = (conv_err || temp_conv_err);
			temp_conv_err = ECC_convertBit2Byte(syn, ham.r, &syn_byte);
			conv_err = (conv_err || temp_conv_err);

			if ((mat_pos == syn_byte) && (conv_err == false)) {
				pos = row;

				// printf("err pos: %u\n", pos);
				// correct the incorrect bit
				if (data[pos] == 1) {
					data[pos] = 0;

				} else if (data[pos] == 0) {
					data[pos] = 1;

				} else {
					printf("Bit not readable");
					ECC_printBits(&data[pos]);
				}
				// fill the err stat
				if (COUNT_ERR == true) {
					// found error in the block
					ECC_setErrStats(1);
				}
				break;
			} else if (conv_err == true) {
				// conversion error
				return;
			}
			row++;
		}
	}
}

uint8_t ECC_dec_DECTED() {

	uint16_t num_encBits;
	num_encBits = ((uint16_t) inBuf_s) * 8;

	uint16_t num_decDataBits;
	num_decDataBits = ((uint16_t) outBuf_s) * 8;

	if ((num_encBits <= ((uint16_t) ECC_MAX_PAYLOAD_SIZE * (uint16_t) 8)) &&
			(num_decDataBits <= ((uint16_t) ECC_MAX_DEC_PAYLOAD_SIZE * (uint16_t) 8))) {
		ECC_doHammingDecoding(inBuf, inBuf_s, outBuf, outBuf_s);

	} else {
		printf("data for decoding too large\t");
	}
	return outBuf_s;
}



void ECC_doDECTEDCorrection(uint8_t* data, uint8_t data_s, uint8_t* syn) {

	uint8_t err_code = 0;
	bool conv_err;
	conv_err = ECC_convertBit2Byte(syn, ham.r, &err_code);
	uint8_t num_err = 0;

	if ((err_code > 0) && (conv_err == false)) {
		// check if error is a single bit error
		uint8_t i;
		for (i = 0; i < sizeof(singleBitError_patterns_DECTED); i++) {
			if (err_code == singleBitError_patterns_DECTED[i]) {
				// case single bit error in whole encoded bits
				// swap bit
				ECC_swapBit(data, data_s, i);
				// save the num of corrected errors
				num_err = 1;
				break;
			}
		}
		uint8_t row, col;
		// check if error is a double bit error in the data bits or the parity block
		for (row = 0; row < sizeof(data2BitErrorSameBlock_patterns_DECTED[0]); row++) {
			for (col = 0; col < sizeof(data2BitErrorSameBlock_patterns_DECTED[0]); col++) {
				if (err_code == data2BitErrorSameBlock_patterns_DECTED[row][col]) {
					if (col > row) {
						// case double error in data bits
						// swap bit
						ECC_swapBit((uint8_t*) &data[0], 8, row);
						ECC_swapBit((uint8_t*) &data[0], 8, col);
						// save the num of corrected errors
						num_err = 2;
						break;
					} else if (col < row) {
						// case double error in parity bits
						// swap bit
						ECC_swapBit((uint8_t*) &data[8], 8, row);
						ECC_swapBit((uint8_t*) &data[8], 8, col);
						// save the num of corrected errors
						num_err = 2;
						break;
					}
				}
			}
		}

		// check if error is a double bit error in the parity and data section
		for (row = 0; row < sizeof(parityAndData2BitError_patterns_DECTED[0]); row++) {
			for (col = 0; col < sizeof(parityAndData2BitError_patterns_DECTED[0]); col++) {

				if (err_code == parityAndData2BitError_patterns_DECTED[row][col]) {
					// swap bit
					ECC_swapBit((uint8_t*) &data[0], 8, row); // error correction in data bits
					ECC_swapBit((uint8_t*) &data[8], 8, col); // error correction in parity bits
					// save the num of corrected errors
					num_err = 2;
					break;
				}
			}
		}
		// fill the err stat
		if (COUNT_ERR == true) {
			// found error in the block
			ECC_setErrStats(num_err);
		}
	} else if (conv_err == true) {
		// conversion error
		return;
	}
}


bool ECC_enc_bch() {

	// protect encoding process
	isEncoding = true;

	// reset bytes of output
	char outBuf_work[outBuf_s];
	memset(outBuf_work, 0, outBuf_s);

	// calculate size
	uint8_t tempBuf_s;
	tempBuf_s = inBuf_s / current_payload_bytes;

	if ((inBuf_s % current_payload_bytes) != 0) {
		tempBuf_s += 1;
	}
	uint8_t numOfBlck = tempBuf_s;

	tempBuf_s *= current_payload_bytes;

	// reset buffer since not all bytes are written with data (need for an array since otherwise memset does not work)
	uint8_t tempBuf[tempBuf_s];
	memset(tempBuf, 0, tempBuf_s);


	// copy the data to send to the working buffer
	memcpy(tempBuf, inBuf, inBuf_s);

	uint8_t blck_counter = 1;
	// split data into blocks
	uint8_t k = 0;
	uint8_t i = 0;
	for (k = 0; k < tempBuf_s; k = k + current_payload_bytes) {

		// do the block encoding and add size info if necessary in the last block
		if (blck_counter == numOfBlck) {
			uint8_t block_size = inBuf_s - ((numOfBlck -1) * current_payload_bytes);
			ECC_doDivisionEncoding((char *) &tempBuf[k], (char *) &outBuf_work[i], block_size);
		} else {
			ECC_doDivisionEncoding((char *) &tempBuf[k], (char *) &outBuf_work[i], current_payload_bytes);
		}

		i += 8;
		blck_counter++;
	}

	// copies back encoded data to the output buffer
	memcpy(outBuf, (char *) &outBuf_work, outBuf_s);

	// protect encoding process finished
	isEncoding = false;

	return true;
}

void ECC_printBytes(char* buf, uint8_t buf_s) {
	uint8_t pos;
	if (buf_s > 0) {
		printf("---- buffer start ----\n");
		printf("%-8s", "bits"); printf(" | ");
		printf("%-8s", "hex"); printf(" | ");
		printf("%-8s", "uint8_t"); printf(" | ");
		printf("%-8s", "addr"); printf("\n");
		for (pos = 0; pos < buf_s; pos++) {
			ECC_printBits(&buf[pos]);
		}
		printf("---- buffer end ----\n");
	}
}


void ECC_doDivisionEncoding(char* dataBytes, char* enc_dataBytes, uint8_t usedPayloadBytes){



	uint8_t data[bch.k];
	memset((uint8_t *) &data, 0, bch.k);

	//printf("used payload bytes in block: %d\n", usedPayloadBytes);

	uint16_t dataBytes_bits = (uint16_t) current_payload_bytes * 8;
	// data in array form
	ECC_convertBytes2Bits(dataBytes, current_payload_bytes, (uint8_t *) &data, dataBytes_bits);
	//printf("data\n");
	//ECC_printBytes((uint8_t *) &data, bch.k);

	// include size info about num of effectively used bytes in the block if necessary$
	if (usedPayloadBytes < current_payload_bytes) {
		//printf("inserting 1 bit @%d\n", bch.k);
		data[bch.k - 1] = 1;
		ECC_setSizeBits((uint8_t*) &data[bch.k - 2], usedPayloadBytes);
	} else {
		// block is full
		data[bch.k - 1] = 0;
	}

	uint8_t rem_ar[bch.n - bch.k];
	ECC_doPolyDiv((uint8_t*) &data, (uint8_t*) &rem_ar);
	//printf("reminder poly bytes:\n");
	//ECC_printBytesRow((uint8_t *) &rem_ar, bch.n - bch.k);

	//printf("data\n");
	//ECC_printBytes((uint8_t *) &data, bch.k);

	// convert data back to bitstream

	unsigned long long shifted_data = 0;
	//	ECC_convertBits2Bytes((uint8_t *) &data, (current_payload_bytes * 8) + 1, (uint8_t*) &shifted_data);
	ECC_convertBits2Bytes((uint8_t *) &data, bch.k, (uint8_t*) &shifted_data, 8);

	//printf("data\n");
	//ECC_printBytesRow((uint8_t*) &shifted_data, 8);

	// shift data to make space for reminder polynome
	//printf("shifted data\n");
	ECC_lshift64((uint8_t *) &shifted_data, current_gen_poly_deg);
	//ECC_printBytesRow((uint8_t *) &shifted_data, 8);

	// convert reminder poly to bitstream
	//printf("rem\n");
	unsigned long long rem = 0;
	ECC_convertBits2Bytes((uint8_t *) &rem_ar, bch.n - bch.k, (uint8_t*) &rem, 8);
	//ECC_printBytesRow((uint8_t *) &rem, 8);

	// append reminder bits to shifted data
	ECC_xor64((char *) &shifted_data, (char*) &rem, (char *) &shifted_data);

	// split data into bytes
	memcpy(enc_dataBytes, (char *) &shifted_data, 8);
}


void ECC_convertBits2Bytes(uint8_t* in, uint16_t in_s, uint8_t* out, uint8_t out_s) {
	// transform the encoded bits to bytes
	uint8_t v, w;
	v = 0;
	uint16_t num_bytes = 0;
	if (in_s % 8 != 0) {
		num_bytes = (in_s / 8) + 1;
	} else {
		num_bytes = in_s / 8;
	}

	if ((in_s > ((uint16_t) ECC_MAX_PAYLOAD_SIZE * (uint16_t) 8)) ||
			(out_s > ECC_MAX_PAYLOAD_SIZE) ||
			(num_bytes > ECC_MAX_PAYLOAD_SIZE)) {
		printf("size error in bits-2-bytes converter\t");
		return;
	}

	uint8_t byte = 0;
	uint8_t bitpos = 0;
	uint16_t u;
	for (u = 0, w = 0; (u < in_s); u++) {
		if (in[u] == 1) {
			byte += potencies[bitpos];
			out[w] = byte;
		}
		if (bitpos == 7) {
			w++;
			bitpos = 0;
			byte = 0;
		} else {
			bitpos++;
		}
	}
}


void ECC_doPolyDiv(uint8_t* data, uint8_t* rem) {

	int i, j ;
	uint8_t feedback = 0;

	for (i = 0; i < bch.n - bch.k; i++) {
		// reset reminder bits
		rem[i] = 0;
	}
	for (i = bch.k - 1; i >= 0; i--) {
		feedback = (data[i] ^ rem[bch.n - bch.k - 1]);
		if (feedback != 0) {
			for (j = bch.n - bch.k - 1; j > 0; j--) {
				if (current_gen_poly[j] != 0)
					rem[j] = (rem[j - 1] ^ feedback);
				else
					rem[j] = rem[j - 1];
			}
			rem[0] = (current_gen_poly[0] && feedback);
		} else {
			for (j = bch.n - bch.k - 1; j > 0; j--){
				rem[j] = rem[j - 1];
			}
			rem[0] = 0;
		}
	}
}


void ECC_setSizeBits(uint8_t* bit_ptr, uint8_t num_bytes_used) {
	if (num_bytes_used > 6) {
		printf("wrong num of payload bytes\n");
	}
	else {
		// reset the bits
		*bit_ptr = 0;
		*(bit_ptr - 1) = 0;
		*(bit_ptr - 2) = 0;

		uint8_t offset = 0x01;

		uint8_t i;
		for (i = 0; i < 3; i++) {
			if ((num_bytes_used & offset) == offset) {
				*bit_ptr = 1;
			}
			offset = (offset << 1);
			bit_ptr--;
		}
	}
}


void ECC_convertBytes2Bits(char* inBuf, uint8_t inBuf_s, char* outBuf, uint16_t out_s) {
	uint8_t pos;
	uint8_t bitPos;
	uint8_t offset;

	if ((inBuf_s > ECC_MAX_PAYLOAD_SIZE) ||
			(out_s > ((uint16_t) ECC_MAX_PAYLOAD_SIZE * (uint16_t) 8))
		) {
		printf("size error in bytes-2-bits converter\t");
		return;
	}

	int bitCount;
	bitCount = 0;
	for (pos = 0; pos < inBuf_s; pos++) {
		offset = 0x01;
		// Goes over every bit of a byte
		for (bitPos = 0; bitPos < 8; bitPos++) {
			if (((char) inBuf[pos] & offset) == offset){
				outBuf[bitCount] = 1;
			} else {
				outBuf[bitCount] = 0;
			}
			offset = (offset << 1);
			bitCount++;
		}
	}
}

uint8_t ECC_dec_bch() {

	bool hasError = false;
	// count the errors corrected for stats
	uint8_t numCorrErr = 0;

	// split data into blocks of 8 bytes
	// calculate number of encoded blocks
	uint8_t num_blocks = inBuf_s / 8;

	if (inBuf_s % 8 != 0) {
		printf("ECC_dec_bch: size does not correspond\t");
		num_blocks += 1;
	}

	long long dataBlock_rcv = 0;
	// used for comparison
	long long temp_dataBlock_rcv = 0;

	uint8_t error_bits[current_err_capability];

	// current number of payload bytes in already inserted
	uint8_t current_payload_len = 0;

	// do decoding and correction for each block
	uint8_t block;
	for (block = 0; block < num_blocks; block++) {
		// count the errors for every block
		numCorrErr = 0;

		// do polynom division to see if there's an error occurring
		//long long dataBlock_rcv = inBuf[block * 8];
		dataBlock_rcv = 0;
		temp_dataBlock_rcv = 0;
		memcpy((uint8_t*) &dataBlock_rcv, (uint8_t*) &inBuf[block * 8], 8);

		// always set the 63. bit to zero. this bit is never touched by the encoder but needs to be transmitted in order to fit in 8 bytes
		unsigned long long res_bit = 0x7FFFFFFFFFFFFFFF;

		ECC_and64((char*) &dataBlock_rcv, (char*) &res_bit, (char*) &dataBlock_rcv);

		if (COUNT_ERR == true) {
			// copy the original data to the temp arry for comparison
			memcpy((uint8_t*) &temp_dataBlock_rcv, (uint8_t*) &dataBlock_rcv, 8);
		}

		// syndrome calculation
		uint8_t synd[2 * current_err_capability];
		hasError = ECC_calcSyndromes((uint8_t *) &dataBlock_rcv, (uint8_t *) &synd);

		if (hasError) {
//			printf("Error detected\n");
			// syndrome error location calculation using Berlekamp-Massey Algorithm (optimized for binary)
//			printf("Berlekamp-Massey\n");
			int err_poly[2 * current_err_capability];
			// reset error poly
			uint8_t p;
			for (p = 0; p < 2 * current_err_capability; p++) {
				err_poly[p] = -1;
			}
			ECC_doBerleMass((uint8_t *) &synd, (int*) &err_poly);
			// chien search to find error positions
//			printf("Chien\n");
			memset((uint8_t*) &error_bits, 0, current_err_capability);
			numCorrErr = ECC_doChienSearch((int*) &err_poly, (uint8_t*) error_bits);
			// correct bit errors
			long long offset = 0x0000000000000001;
			uint8_t bit;
			for (bit = 0; bit < current_err_capability; bit++) {
				offset = 0x0000000000000001;
				offset = (offset << error_bits[bit]);
				ECC_xor64((char*) &dataBlock_rcv, (char*) &offset, (char*) &dataBlock_rcv);
			}
		} else {
//			printf("No error occurred\n");
		}
		// find the number of payload bits to read

		// chop off the reminder bits
		ECC_rshift64((char*) &dataBlock_rcv, current_gen_poly_deg);

		// find the relevant data in the stream in order to correctly merge the blocks together
		uint8_t num_payload_bytes = ECC_readSizeBits((char*) &dataBlock_rcv);

		// copy back the decoded and corrected bit stream
		memcpy((uint8_t*) &outBuf[current_payload_len], (uint8_t*) &dataBlock_rcv, num_payload_bytes);

		// add the already num of bytes
		current_payload_len += num_payload_bytes;

		// set the error log
		ECC_setErrStats(numCorrErr);
	}
	return current_payload_len;
}


int ECC_readSizeBits(uint8_t* data) {

	int num = 0;
	unsigned long long offset = 0x0000000000000001;
	ECC_lshift64((char*) &offset, bch.k - 1);

	unsigned long long res = 0;
	ECC_and64(data, (uint8_t *) &offset, (uint8_t *) &res);

	// find if there is a size info encoded
	if (res == offset) {
		// reset the full / notfull bit
		ECC_xor64(data, (uint8_t *) &offset, data);
		uint8_t i;
		for (i = 0; i < 3; i++) {
			ECC_rshift64((char*) &offset, 1);
			ECC_and64(data, (uint8_t *) &offset, (uint8_t *) &res);
			if (res == offset) {
				if (i == 0) {
					num += 1;
				} else if (i == 1) {
					num += 2;
				} else if (i == 2){
					num += 4;
				}
				// reset the read bit
				ECC_xor64(data, (uint8_t *) &offset, data);
			}
		}
		if ((num > 0) && (num <= 6)) {
			return num;
		} else {
			return 0;
		}
	} else {
		return current_payload_bytes;
	}
}


void ECC_rshift64(char* l1, uint8_t shift) {
	// mult and div with 2 works since there's no overflow (no cyclic overflow)
	unsigned long long temp = 0;
	memcpy((uint8_t*) &temp, l1, 8);
	uint8_t i;
	for (i = 0; i < shift; i++) {
			// right shift = / 2
			temp /= 2;
	}
	// copy back the result
	memcpy(l1, (uint8_t*) &temp, 8);
}

void ECC_lshift64(char* l1, uint8_t shift) {
	// mult and div with 2 works since there's no overflow (no cyclic overflow)
	unsigned long long temp = 0;
	memcpy((uint8_t*) &temp, l1, 8);
	uint8_t i;
	for (i = 0; i < shift; i++) {
			// left shift = * 2
			temp *= 2;
	}
	// copy back the result
	memcpy(l1, (uint8_t*) &temp, 8);
}

void ECC_xor64(char* l1, char* l2, char* res) {

	uint16_t long1[4];
	uint16_t long2[4];
	uint16_t result[4] = {0, 0, 0, 0};

	memcpy((char*) &long1, l1, 8);
	memcpy((char*) &long2, l2, 8);

	uint8_t k;
	for (k = 0; k < 4; k++) {
		result[k] = (long1[k] ^ long2[k]);
	}
	// write result to target
	memcpy(res, (uint8_t*) &result, 8);

}


uint8_t ECC_doChienSearch(int* err_poly, uint8_t* error_bits) {
	// root counter used to determine if an uncorrectable but detectable error has happen
	uint8_t root_count = 0;
	// holds the roots found
	uint8_t roots[2 * current_err_capability];
	// reset the roots
	memset((uint8_t*) &roots, 255, 2 * current_err_capability);
	uint8_t root_pos = 0; // counts the error
	// needed to find out the degree of the error polynome
	uint8_t deg_err_poly = 0;

	uint8_t i;
	for (i = 0; i <= 62; i++) {
		int k;
		uint8_t result = 1;
		for (k = 1; k < 2 * current_err_capability; k++) {
			// goes over polynome testing every power of alpha
			if (err_poly[k] != -1) {
				// adds the coefficient in the error poly and the replaces x (with inserted alpha power)
				result ^= alpha_to[(err_poly[k] + (k * i)) % 63];
				// finds the degree of the error polynome
				deg_err_poly = k;
			}
		}
		if (result == 0) {
			// found root -> increment counter
			root_count++;
			// test if root is distinct
			if (!ECC_findRoot((uint8_t*) &roots, i)) {
				// the roots are the inverse error positions
				roots[root_pos] = i;
				// map errors to bit position
				error_bits[root_pos] = 63 - i;
				root_pos++;
			}
		}
	}

	// returns the number of errors
	return root_pos;
}

bool ECC_findRoot(uint8_t* roots, uint8_t root) {
	uint8_t k;
	for (k = 0; k < 2 * current_err_capability; k++) {
		if (roots[k] == root) {
			return true;
		}
	}
	return false;
}

void ECC_doBerleMass(uint8_t* synd, int* err_poly) {

	uint8_t L, n;
	L = 0;
	n = 2 * current_err_capability;

	// contains the error polynom
	int c[n];
	int p[n];
	// temporary array t(x)
	int t[n];
	// temporary array used to calc c[x]
	int t2[n];
	// temporary array used to calc d * d_m^-1 * x^l * p(x)
	int t3[n];

	uint8_t g;
	for (g = 0; g < n; g++) {
		c[g] = -1;
		p[g] = -1;
		t[g] = -1;
		t2[g] = -1;
		t3[g] = -1;
	}

	c[0] = 0; // alpha_to[0] = 1
	p[0] = 0; // alpha_to[0] = 1

	// just the loop variable
	uint8_t coef = 0;

	// contains the result of d*d_m^-1
	int factor = -1;
	uint8_t alpha_1;
	uint8_t alpha_2;

	uint8_t l;
	l = 1;
	uint8_t d_m;
	d_m = 1; // alpha_to[0] = 1
	uint8_t d;
	d = 0; // discrepancy

	// start loop over syndromes
	uint8_t k;
	for (k = 1; k < n; k += 2) {
		d = synd[k-1];
		uint8_t i;
		for (i = 1; i <= L; i++) {
			if ((c[i] > -1) && (synd[k-1-i] > 0)){
				d ^= alpha_to[ECC_multGF64(alpha_to[c[i]], synd[k-1-i])];
			}
		}
		if (d == 0) {
			l++;
		} else {
			if ((2*L) >= k) {
				// calcs d*d_m^-1
				factor = (ECC_findAlpha(d) + (63 - ECC_findAlpha(d_m))) % 63;
				// calcs d*d_m^-1*p(x)
				for (coef = 0; coef < n; coef++) {
					if (p[coef] != -1) {
						// alpha mult
						t3[coef] = (factor + p[coef]) % 63;
					} else {
						t3[coef] = -1;
					}
				}

				// calcs d*d_m^-1*p(x)*x^l
				for (coef = 0; coef < n; coef++) {
					t2[(coef + l) % n] = t3[coef];
				}

				// computes the substraction (XOR) of the c(X) and the d*d_m^-1....
				for (coef = 0; coef < n; coef++) {
					if ((c[coef] == -1) && (t2[coef] == -1)) {
						c[coef] = -1;
					} else if (c[coef] == -1) {
						c[coef] = t2[coef];
					} else if ((c[coef] != -1) && (t2[coef] != -1)){
						alpha_1 = alpha_to[c[coef]];
						alpha_2 = alpha_to[t2[coef]];
						if (alpha_1 == alpha_2) {
							c[coef] = -1;
						} else {
							c[coef] = ECC_findAlpha(alpha_1 ^ alpha_2);
						}
					}
				}
				l++;
			} else {
				memcpy((uint8_t*) &t, (uint8_t*) &c, 2 * n);
				// calcs d*d_m^-1
				factor = (ECC_findAlpha(d) + (63 - ECC_findAlpha(d_m))) % 63;

				// calcs d*d_m^-1*p(x)
				for (coef = 0; coef < n; coef++) {
					if (p[coef] != -1) {
						// alpha mult
						t3[coef] = (factor + p[coef]) % 63;
					} else {
						t3[coef] = -1;
					}
				}

				// calcs d*d_m^-1*p(x)*x^l
				for (coef = 0; coef < n; coef++) {
						t2[(coef + l) % n] = t3[coef];
				}

				// computes the substraction (XOR) of the c(X) and the d*d_m^-1....
				for (coef = 0; coef < n; coef++) {

					if ((c[coef] == -1) && (t2[coef] == -1)) {
						c[coef] = -1;
					} else if (c[coef] == -1) {
						c[coef] = t2[coef];
					} else if ((c[coef] != -1) && (t2[coef] != -1)){
						alpha_1 = alpha_to[c[coef]];
						alpha_2 = alpha_to[t2[coef]];
						if (alpha_1 == alpha_2) {
							c[coef] = -1;
						} else {
							c[coef] = ECC_findAlpha(alpha_1 ^ alpha_2);
						}
					}
				}
				L = k - L;
				memcpy((uint8_t*) &p, (uint8_t*) &t, 2 * n);
				d_m = d;
				l = 1;
			}
		}
		l++;
	}
	// write the error polynome
	uint8_t err_poly_coef;
	for (err_poly_coef = 0; err_poly_coef < 2 * current_err_capability; err_poly_coef++) {
		err_poly[err_poly_coef] = c[err_poly_coef];
	}
}




uint8_t ECC_dec_rep(uint8_t r, uint8_t k) {

	uint8_t i,j,b, num1, num0, bytecount;
	// counts the bytes for the output
	bytecount = 0;
	for (i = 0; i < inBuf_s; i += j) {
		uint8_t pos = 0x80;
		uint8_t outByte = 0x00;
		for (b = 0; b < 8; b++) {
			num0 = 0;
			num1 = 0;
			for (j = 0; j < r; j++) {
				if ((inBuf[i+j] & pos) == pos){
					num1++;
				} else {
					num0++;
				}
			}
			// count the error
			if (COUNT_ERR == true) {
				if ((num0 == 0 && num1 == r) || (num0 == r && num1 == 0)) {
					ECC_setErrStats(0);
				} else {
					ECC_setErrStats(1);
				}
			}
			// write the according bit in the byte (Majority logic detection)
			if (num1 > num0) {
				outByte = (outByte | pos);
			}
			// shift pos to right
			pos = (pos >> 1);
		}
		if (bytecount < outBuf_s) {
			// add decoded bytes to the output buffer
			outBuf[bytecount] = outByte;
		} else {
			printf("Decoded data does not fit into buffer - wrong codec?");
			ECC_decoder_state.DECODER_SUCCESSFUL = false;
			return 0;
		}
		bytecount++;
	}
	return bytecount;
}

void ECC_and64(char* l1, char* l2, char* res) {

	uint16_t long1[4];
	uint16_t long2[4];
	uint16_t result[4] = {0, 0, 0, 0};

	memcpy((char*) &long1, l1, 8);
	memcpy((char*) &long2, l2, 8);

	uint8_t k;
	for (k = 0; k < 4; k++) {
		result[k] = (long1[k] & long2[k]);
	}
	// write result to target
	memcpy(res, (uint8_t*) &result, 8);
}

int ECC_multGF64(uint8_t a, uint8_t b) {
	int pow_a = ECC_findAlpha(a);
	int pow_b = ECC_findAlpha(b);
	if ((pow_a >= 0) && (pow_b >= 0)) {
		return (pow_a + pow_b) % 63;
	} else if (pow_a < 0) {
		printf("error in GF64 Mult because of first param\n");
		return -1;
	} else if (pow_b < 0) {
		printf("error in GF64 Mult because of second param\n");
		return -1;
	}
}

int ECC_findAlpha(uint8_t bits) {

	if (bits == 0) {
		// Null is not an element of the field
		printf("0 is not element of GF64\n");
		return -1;
	}
	int pow = -1;
	uint8_t i;
	for (i = 0; i < 64; i++) {
		if (bits == alpha_to[i]) {
			pow = i;
			break;
		}
	}
	if (pow >= 0) {
		return pow;
	} else {
		printf("alpha power %d not found for:\n", pow);
		return -1;
	}
}


bool ECC_calcSyndromes(uint8_t* corr_data_block, uint8_t* synd) {

	// shows if error happened (synd != 0)
	bool hasError = false;

	// number of syndrome values to be calculated
	uint8_t synd_num = 2 * current_err_capability; // corresponds to 2 * t

	uint8_t j;
	for (j = 1; j <= synd_num; j++){
		uint8_t bit;
		// reset syndromes
		synd[j-1] = 0;

		long long offset = 0x0000000000000001;
		long long result = 0;
		uint8_t alpha = 0;

		for (bit = 0; bit < 64; bit++) {
			ECC_and64((uint8_t*) corr_data_block, (uint8_t*) &offset, (uint8_t*) &result);
			if (result == offset) {
				alpha = alpha_to[(bit * j) % 63];
				synd[j-1] = (alpha ^ synd[j-1]);
			}
			offset = (offset << 1);
		}

		if (synd[j-1] != 0) {
			hasError = true;
		}
	}
	return hasError;
}


bool ECC_swapBit(uint8_t* data, uint8_t data_s, uint8_t pos) {
	if (pos < data_s) {
		if (data[pos] == 1) {
			data[pos] = 0;
		} else if (data[pos] == 0) {
			data[pos] = 1;
		} else {
			return false;
		}
		return true;
	} else {
		return false;
	}
}


void ECC_printBits(char* byte) {
	uint8_t j;
	uint8_t pos;
	// pos represents 1 0 0 0 0 0 0 0 which is shifted to right
	pos = 0x80;
	for (j = 0; j < 8; j++) {
		if ((*byte & pos) == pos){
			printf("1");
		} else {
			printf("0");
		}
		// shift pos to right
		pos = (pos >> 1);
	}
	printf(" | 0x");
	printf("%-6x", (uint8_t) *byte);
	printf(" | ");
	printf("%-8u", (uint8_t) *byte);
	printf(" | ");
	printf("%-8p", byte);
	printf("\n");
}



bool ECC_convertBit2Byte(uint8_t* bitseq, uint8_t bitseq_s, uint8_t* res_byte) {

	uint8_t r;
	uint8_t sum;
	sum = 0;
	*res_byte = 0;
	// check that number of bits is max a byte
	if (bitseq_s > 8){
		printf("bits do not fit in one byte (wrong codec? size: %u)\t", bitseq_s);
		printf("res_byte: %u\t", *res_byte);
		return true;
	} else {
		for (r = 0; r < bitseq_s ; r++) {
			if (bitseq[r] == 1) {
				sum += potencies[r];
			}
		}
		if (sum <= 255) {
			*res_byte = sum;
			return false;

		} else {
			printf("size error in bit-2-byte converter\t");
			*res_byte = 0;
			return true;

		}
	}
}


void ECC_setErrStats(uint8_t err) {
	// an additional error num has been added (from a new block)
	ECC_corrected_error_stats.block_counter++;

	if (ECC_corrected_error_stats.block_counter == 1) {
		ECC_corrected_error_stats.max = err;
		ECC_corrected_error_stats.min = err;
	} else {
		if (err > ECC_corrected_error_stats.max) {
			ECC_corrected_error_stats.max = err;
		} else if (err < ECC_corrected_error_stats.min) {
			ECC_corrected_error_stats.min = err;

		}
	}
	// update sum of errors
	ECC_corrected_error_stats.err_sum += (int) err;
}


